/*
 * Copyright (C) 2002  Pekka Enberg <Pekka.Enberg@cs.helsinki.fi>
 *
 * Distributed under the terms of the GNU General Public License
 * version 2 or later.
 */
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Print error message.  */
void error(const char * format, ...)
{
    va_list args;

    fprintf(stderr, "Error: ");

    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);

    fputc('\n', stderr);
}

#define INDENT_CHAR      ' '

#define DEFAULT_NUM_INDENT_CHARS 4
static int num_indent_chars = DEFAULT_NUM_INDENT_CHARS;

/* Print indent characters.  */
void print_indent(FILE * output, int indent_level)
{
    int i;

    for (i = 0; i < (num_indent_chars * indent_level); i++) {
	fputc(INDENT_CHAR, output);
    }
}

/* Buffer used to hold XML stream.  */
struct buffer {
    int * data;
    size_t length;
    size_t capacity;
};

/* Initialize buffer.  */
void buffer_init(struct buffer * buffer, size_t initial_capacity)
{
    buffer->data     = malloc(initial_capacity * sizeof(int));
    buffer->length   = 0;
    buffer->capacity = initial_capacity;
}

/* Release buffer.  */
void buffer_release(struct buffer * buffer)
{
    if (buffer->capacity > 0) {
	free(buffer->data);
	buffer->data     = NULL;
	buffer->capacity = 0;
    }
    buffer->length = 0;
}

/* Check if buffer is full.  */
bool buffer_is_full(struct buffer * buffer)
{
    return (buffer->capacity == buffer->length);
}

/* Grow buffer by one.  */
void buffer_grow(struct buffer * buffer)
{
    buffer->data = realloc(buffer->data, buffer->capacity + 1);
    if (buffer->data == NULL) {
	error("Could not allocate memory for buffer.");
	exit(1);
    }
    buffer->capacity++;
}

/* Push item at back of buffer.  */
void buffer_push(struct buffer * buffer, int item)
{
    if (buffer_is_full(buffer)) {
	buffer_grow(buffer);
    }
    buffer->data[buffer->length] = item;
    buffer->length++;
}

/* Pop item from back of buffer.  */
int buffer_pop(struct buffer * buffer)
{
    int result = -1;

    if (buffer->length > 0) {
	buffer->length--;
	result = buffer->data[buffer->length];
    } else {
	error("Buffer underflow.");
    }
    return result;
}

/* Flush buffer to output stream.  */
void buffer_flush(struct buffer * buffer, FILE * output)
{
    int i;

    for (i = 0; i < buffer->length; i++) {
	fputc(buffer->data[i], output);
    }
    buffer->length = 0;
}

/* Check if for XML begin tag.  */
static inline bool is_xml_begin_tag(int current, int lookahead)
{
    return (('<' == current) && isalpha(lookahead));
}

/* Check for XML end tag.  */
static inline bool is_xml_end_tag(int current, int lookahead)
{
    return ((('<' == current) && ('/' == lookahead)) ||
	    (('/' == current) && ('>' == lookahead)));
}

/* Check for XML comment begin.
   Note: After function call, current and lookahead are not
   valid. Read data is pushed to the buffer.  */
static inline bool is_xml_comment_begin(int current, int lookahead,
					FILE * input, struct buffer * buffer)
{
    bool result = false;

    if (('<' == current) && ('!' == lookahead)) {
	int tmp = fgetc(input);
	if ('-' == tmp) {
	    result = true;
	}
	if (ungetc(tmp, input) != tmp) {
	    error("Could not restore XML stream.");
	}
    }
    return result;
}

/* Check for XML comment end.  */
static inline bool is_xml_comment_end(int current, int lookahead,
				      struct buffer * buffer)
{
    bool result = false;

    if (('-' == current) && ('>' == lookahead)) {
	int tmp = buffer_pop(buffer);

	if ('-' == tmp) {
	    result = true;
	}
	buffer_push(buffer, tmp);
    }
    return result;
}

#define BUFFER_INITIAL_CAPACITY 1024

/* Indent XML stream.  */
void indent(FILE * input, FILE * output)
{
    struct buffer buffer;

    int indent_level;
    int lookahead;
    int begin_tag_count;
    bool eat_whitespace;
    bool inside_comment;

    buffer_init(&buffer, BUFFER_INITIAL_CAPACITY);

    indent_level    = 0;
    lookahead       = fgetc(input);
    begin_tag_count = 0;
    eat_whitespace  = true;
    inside_comment  = false;

    for (;;) {
	int current = lookahead;

	if (EOF == current) {
	    break;
	}
	lookahead = fgetc(input);

	/* Eat characters until non-whitespace character is found.  */
	if (eat_whitespace) {
	    if (isspace(current)) {
		continue;
	    } else {
		eat_whitespace = false;
	    }
	}

	if (!inside_comment) {
	    if (is_xml_begin_tag(current, lookahead)) {
		/* XML begin tag changes indent level only after we have
		   finished the current line.  */
		begin_tag_count++;
	    } else if (is_xml_end_tag(current, lookahead)) {
		if (begin_tag_count > 0) {
		    /* Begin and end tag on the same line balance each
                       other.  */
		    begin_tag_count--;
		} else {
		    indent_level--;
		    if (indent_level < 0) {
			error("XML is not well-formed.");
			indent_level = 0;
		    }
		}
	    }
	}

	if (is_xml_comment_begin(current, lookahead, input, &buffer)) {
	    inside_comment = true;
	} else if (is_xml_comment_end(current, lookahead, &buffer)) {
	    inside_comment = false;
	}

	buffer_push(&buffer, current);

	if ('\n' == current) {
	    print_indent(output, indent_level);
	    buffer_flush(&buffer, output);

	    if (begin_tag_count > 0) {
		indent_level += begin_tag_count;
		begin_tag_count = 0;
	    }
	    eat_whitespace = true;
	}
    }

    buffer_release(&buffer);
}

int main(int argc, char * argv[])
{
    FILE * input;

    if (argc > 1) {
	input = fopen(argv[1], "r");
	if (NULL == input) {
	    error("Could not open '%s': %s.", argv[1], strerror(errno));
	    return EXIT_FAILURE;
	}
    } else {
	input = stdin;
    }

    indent(input, stdout);
    fclose(input);

    return EXIT_SUCCESS;
}
